#include "Rainbow.h"
#include <Wire.h>
#include <avr/pgmspace.h>

//=============================================================
extern unsigned char dots_color[2][3][8][4];  //define Two Buffs (one for Display ,the other for receive data)
extern unsigned char GamaTab[16];             //define the Gamma value for correct the different LED matrix
extern unsigned char Prefabnicatel[5][3][8][4];  //define preprogrammed pictures
extern unsigned char ASCII_Char[52][8];
extern unsigned char ASCII_Number[10][8];

extern unsigned char RainbowCMD[4][32];  //the glorious command structure array
//=============================================================
unsigned char Address=4;  //Address of the device. Note: this must be changed and compiled for all unique Rainbowduinos
unsigned char line,level;
unsigned char Buffprt=0;  //define variable that keeps track of which buffer is displaying and which recieves
unsigned char State=0;  //the state of the Rainbowduino, used to figure out what to do next
unsigned char g8Flag1;  //flag for onrequest from IIC to if master has asked
unsigned char cmdsession=0;  //the number of the cmdsession that last took place, total number that has occured, first is 1 not 0


void setup()
{
  _init();

}

void loop()
{
  switch (State)
  {
  case waitingcmd:   
    break;

  case morecmd:
    break;

  case processing:
    RunCMD();
    State=checking;
    break;

  case checking:
    if(CheckRequest)
    {
      State=waitingcmd;
      ClrRequest;
    }
    break;

  default:
    State=waitingcmd; 
    break;
  }

}

void _init(void)    // define the pin mode
{
  DDRD=0xff;
  DDRC=0xff;
  DDRB=0xff;
  PORTD=0;
  PORTB=0;
  Wire.begin(Address); // join i2c bus (address optional for master) 
  Wire.onReceive(receiveEvent); // define the receive function for receiving data from master
  Wire.onRequest(requestEvent); // define the request function for the request from maseter 
  init_timer2();  // initialize the timer for scanning the LED matrix
}

//============INTERRUPTS======================================

ISR(TIMER2_OVF_vect)          //Timer2  Service 
{ 
  TCNT2 = GamaTab[level];    // Reset a  scanning time by gamma value table
  flash_next_line(line,level);  // sacan the next line in LED matrix level by level.
  line++;
  if(line>7)        // when have scaned all LEC the back to line 0 and add the level
  {
    line=0;
    level++;
    if(level>15)       level=0;
  }
}

void init_timer2(void)               
{
  TCCR2A |= (1 << WGM21) | (1 << WGM20);   
  TCCR2B |= (1<<CS22);   // by clk/64
  TCCR2B &= ~((1<<CS21) | (1<<CS20));   // by clk/64
  TCCR2B &= ~((1<<WGM21) | (1<<WGM20));   // Use normal mode
  ASSR |= (0<<AS2);       // Use internal clock - external clock not used in Arduino
  TIMSK2 |= (1<<TOIE2) | (0<<OCIE2B);   //Timer2 Overflow Interrupt Enable
  TCNT2 = GamaTab[0];
  sei();   
}


//=============HANDLERS======================================

void receiveEvent(int howMany)
{
  unsigned char i=0;
  while(Wire.available()>0)
  { 
    RainbowCMD[cmdsession][i]=Wire.receive();  //recieve whatever is available
    i++;
  }
  cmdsession++;  //increment that the session has finished

  switch (RainbowCMD[0][0]){

  case 'F':
    if(cmdsession<4) State=morecmd;  //if not all 4 have been recieved change state to ask for more

    else if ((i==32)&&(cmdsession==4)){  //if 4th session occured, and double check it's the right size
      State=processing;
      cmdsession=0;  //reset cmdsession
    }
    break;
    
  case 'T':
    if((i==30)&&(cmdsession==1)){  //if the correct number of bytes have been sent
      State=processing;
      cmdsession=0;  //reset
    }
    break;      
  
  case 'D':
    if((i==20)&&(cmdsession==1)){  //if the correct number of bytes have been sent
      State=processing;
      cmdsession=0;  //reset
    }
    break;

  case 'M':
    if((i==11)&&(cmdsession==1)){  //if the correct number of bytes have been sent
      State=processing;
      cmdsession=0;  //reset
    }
    break;

  case 'R':
    if((i==5)&&(cmdsession==1)){  //if the correct number of bytes have been sent
      State=processing;
      cmdsession=0;  //reset
    }
    break;
  }
}

void requestEvent(void)  //when the master requests from the slave, this is the handler
{
  unsigned char trans;  //what is to be transmitted

  if (State==morecmd) trans=(morecmd|cmdsession);  //when asking for more data, send the number of the last session recived, then master knows to send the next one
  else trans=State;  //otherwise tell the master what state the slave is in

  Wire.send(trans); 

  if ((State==processing)||(State==checking))  SetRequest;  //the slave is now working on a command, can continue to run without receiving data
}

//==============DISPSHOW========================================
void DispshowFrame(void)
{
  unsigned char color,row,dots,correctcol;

  for(color=0;color<4;color++)
  {
    switch (color) //fixes the fact that the buffer needs GRB, not RGB
    {
    case 0:  //in frame Red
      correctcol = 2;  //out green
      break;
    case 1:  //in frame Green
      correctcol = 1; //out red
      break;
    case 2:  //in frame Blue
      correctcol = 3; //out blue
      break;
    }
    for (row=0;row<8;row++)
    {
      for (dots=0;dots<4;dots++)
      {
        dots_color[((Buffprt+1)&1)][color][row][dots]=RainbowCMD[correctcol][(4*row)+dots];  //get byte info for two dots directly from command
      }
    }
  }
  Buffprt++;  //increment buffer, will switch which one reads from and other writes to
  Buffprt&=1;
}


//===============================================================
void DispshowMono(void)
{
  unsigned char Col_Red,Col_Blue,Col_Green;
  unsigned char color,row,dots,correctcol,tempword,tempdata;

  Col_Red=(RainbowCMD[0][1]&0x0F);  //get the color from command
  Col_Green=((RainbowCMD[0][2]>>4)&0x0F);
  Col_Blue=(RainbowCMD[0][2]&0x0F);

  for(color=0;color<3;color++)
  {
    switch (color){
      case 0:
        correctcol=Col_Green;
        break;
      case 1:
        correctcol=Col_Red;
        break;
      case 2:
        correctcol=Col_Blue;
        break;
    }

    for (row=0;row<8;row++){	
      tempword=RainbowCMD[0][row+3];  //a tempword is a row of on/off store as a byte

      for (dots=0;dots<4;dots++){
        if((tempword<<(2*dots))&0x80){  //if theres a value (1) for left led
          tempdata&=0x0F;  //purge previously store info for those 4 bits
          tempdata|=(correctcol<<4);  //write the new ones for the left led
        }
        else{
          tempdata&=0x0F;  //otherwise just purge
        }
        if((tempword<<(2*dots+1))&0x80){  // there is a value for the right led
          tempdata&=0xF0;  //purge previous
          tempdata|=correctcol; //write new
        }
        else{
          tempdata&=0xF0;  //just purge
        }   

        dots_color[((Buffprt+1)&1)][color][row][dots]=tempdata;  //write the new byte(left/right) info to buffer
      }
    }
  }
  Buffprt++;  //increment buffer
  Buffprt&=1;
}


//===============================================================
void DispshowDich(void)
{
  unsigned char Red1,Red2,Green1,Green2,Blue1,Blue2;
  unsigned char color,row,dots,tempdata;
  unsigned char correctcol1,correctcol2,tempword1,tempword2;

  Red1=((RainbowCMD[0][1]>>4)&0x0F);  //get colors
  Green1=(RainbowCMD[0][1]&0x0F);
  Blue1=((RainbowCMD[0][2]>>4)&0x0F);
  Red2=(RainbowCMD[0][2]&0x0F);
  Green2=((RainbowCMD[0][3]>>4)&0x0F);
  Blue2=(RainbowCMD[0][3]&0x0F);

  for(color=0;color<3;color++)
  {
    switch (color){  //do color correction
      case 0:
        correctcol1=Green1;
        correctcol2=Green2;
        break;
      case 1:
        correctcol1=Red1;
        correctcol2=Red2;
        break;
      case 2:
        correctcol1=Blue1;
        correctcol2=Blue2;
        break;
    }

    for (row=0;row<8;row++){	
      tempword1=RainbowCMD[0][row+4];  //get two tempwords, one for first and second color
      tempword2=RainbowCMD[0][row+12];

      for (dots=0;dots<4;dots++){
        
        if((tempword1<<(2*dots))&0x80){  //left pixel live w/ first color
          tempdata&=0x0F;  //purge previous
          tempdata|=(correctcol1<<4);  //write new color
        }
        else if((tempword2<<(2*dots))&0x80){  //left pixel live with second color
          tempdata&=0x0F;  //purge previous
          tempdata|=(correctcol2<<4);  //write new colo
        }
        else{
          tempdata&=0x0F;  // if not live purge
        }
        
        if((tempword1<<(2*dots+1))&0x80){  //right pixel live w/ first color
          tempdata&=0xF0;  //purge previous
          tempdata|=correctcol1;  //write new color
        }
        else if((tempword2<<(2*dots+1))&0x80){  //right pixel live w/ second color
          tempdata&=0xF0;  //purge previous
          tempdata|=correctcol2;  //write new color
        }
        else{
          tempdata&=0xF0;  //if not live purge
        }   

        dots_color[((Buffprt+1)&1)][color][row][dots]=tempdata;	  
      }
    }
  }
  Buffprt++;
  Buffprt&=1;
}


//===============================================================
void DispshowTrich(void)
{
  unsigned char Red1,Red2,Red3,Green1,Green2,Green3,Blue1,Blue2,Blue3;
  unsigned char color,row,dots,tempdata;
  unsigned char correctcol1,correctcol2,correctcol3,tempword1,tempword2,tempword3;

  Red1=((RainbowCMD[0][1]>>4)&0x0F);  //more color data!
  Green1=(RainbowCMD[0][1]&0x0F);
  Blue1=((RainbowCMD[0][2]>>4)&0x0F);
  Red2=(RainbowCMD[0][2]&0x0F);
  Green2=((RainbowCMD[0][3]>>4)&0x0F);
  Blue2=(RainbowCMD[0][3]&0x0F);
  Red3=(RainbowCMD[0][4]&0x0F);
  Green3=((RainbowCMD[0][5]>>4)&0x0F);
  Blue3=(RainbowCMD[0][5]&0x0F);
  
  for(color=0;color<3;color++)
  {
    switch (color){  //correction like before
      case 0:
        correctcol1=Green1;
        correctcol2=Green2;
        correctcol3=Green3;
        break;
      case 1:
        correctcol1=Red1;
        correctcol2=Red2;
        correctcol3=Red3;
        break;
      case 2:
        correctcol1=Blue1;
        correctcol2=Blue2;
        correctcol3=Blue3;
        break;
    }

    for (row=0;row<8;row++){	
      tempword1=RainbowCMD[0][row+6];  //more words like before, now 3 of them
      tempword2=RainbowCMD[0][row+14];
      tempword3=RainbowCMD[0][row+22];

      for (dots=0;dots<4;dots++){
        
        if((tempword1<<(2*dots))&0x80){  //left pixel live w/ first color
          tempdata&=0x0F;  //purge previous
          tempdata|=(correctcol1<<4);  //write new color
        }
        else if((tempword2<<(2*dots))&0x80){  //left pixel live with second color
          tempdata&=0x0F;  //purge previous
          tempdata|=(correctcol2<<4);  //write new color
        }
        else if((tempword3<<(2*dots))&0x80){  //left pixel live with third color
          tempdata&=0x0F;  //purge previous
          tempdata|=(correctcol3<<4);  //write new color
        }
        else{
          tempdata&=0x0F;  // if not live purge
        }
        
        if((tempword1<<(2*dots+1))&0x80){  //right pixel live w/ first color
          tempdata&=0xF0;  //purge previous
          tempdata|=correctcol1;  //write new color
        }
        else if((tempword2<<(2*dots+1))&0x80){  //right pixel live w/ second color
          tempdata&=0xF0;  //purge previous
          tempdata|=correctcol2;  //write new color
        }
        else if((tempword3<<(2*dots+1))&0x80){  //right pixel live w/ third color
          tempdata&=0xF0;  //purge previous
          tempdata|=correctcol3;  //write new color
        }
        else{
          tempdata&=0xF0;  //if not live purge
        }   

        dots_color[((Buffprt+1)&1)][color][row][dots]=tempdata;	  
      }
    }
  }
  Buffprt++;
  Buffprt&=1;
}

//===============================================================
void DispshowColor(void)
{   
  unsigned char color=0,row=0,dots=0;
  unsigned char Gr,Bl,Re;

  Re=(RainbowCMD[0][2]&0x0F);  //colors! Yay!
  Gr=((RainbowCMD[0][3]>>4)&0x0F);
  Bl=(RainbowCMD[0][3]&0x0F);
  RainbowCMD[0][1]=0;

  for(color=0;color<3;color++)
  {
    for (row=0;row<8;row++)
    {
      for (dots=0;dots<4;dots++)
      {
        switch (color)
        {
        case 0://green
          dots_color[((Buffprt+1)&1)][color][row][dots]=( Gr|(Gr<<4));  //set left and right led to color value
          break;

        case 1://red
          dots_color[((Buffprt+1)&1)][color][row][dots]= (Re|(Re<<4));
          break;

        case 2://blue
          dots_color[((Buffprt+1)&1)][color][row][dots]= (Bl|(Bl<<4));
          break;

        default:
          break;
        }
      }
    }
  }
  Buffprt++;
  Buffprt&=1;
}
//=============================================================
void DispshowPicture(void)
{
  unsigned char pi,shifts;
  unsigned char color=0,row=0,dots=0;
  unsigned char temp;
  unsigned char fir,sec;

  shifts=((RainbowCMD[0][2]>>4)&0x0F);
  pi=RainbowCMD[0][4];
  RainbowCMD[0][1]=0;

  for(color=0;color<3;color++)
  {
    for (row=0;row<8;row++)
    {
      for (dots=0;dots<4;dots++)
      {

        if (shifts&0x01)
        {             
          temp = dots + (shifts>>1);
          fir=pgm_read_byte(&(Prefabnicatel[pi][color][row][(temp<4)?(temp):(temp-4)]));  //magic shifting math
          sec=pgm_read_byte(&(Prefabnicatel[pi][color][row][(temp<3)?(temp+1):(temp-3)]));  //more
          dots_color[((Buffprt+1)&1)][color][row][dots] = (fir<<4)|(sec>>4);    //write to buffer
        }        
        else
        {
          temp = dots + (shifts>>1);
          dots_color[((Buffprt+1)&1)][color][row][dots] = pgm_read_byte(&(Prefabnicatel[pi][color][row][(temp<4)?(temp):(temp-4)]));  //more magic, I didn't code this and i don't get it!
        }
      }
    }
  }
  Buffprt++;
  Buffprt&=1;
}

//===============================================================
void DispshowChar(void)
{
  unsigned char Col_Red,Col_Blue,Col_Green,shift,ASCII;
  unsigned char tempword,color,row,dots,Num,tempdata,correctcol,AS;

  shift=((RainbowCMD[0][2]>>4)&0x0F);  //parse out important stuff from bytes of command
  Col_Red=(RainbowCMD[0][2]&0x0F);
  Col_Green=((RainbowCMD[0][3]>>4)&0x0F);
  Col_Blue=(RainbowCMD[0][3]&0x0F);
  ASCII=RainbowCMD[0][4];
  RainbowCMD[0][1]=0;


  if((ASCII>64)&&(ASCII<91)) AS=ASCII-65;  //ascii are essential a number 0-255 corresponding to a character,
  else if((ASCII>96)&&(ASCII<123)) AS=ASCII-71;  //this 'recenters' them to start from 0 to read from the storage array
  else if( (ASCII>='0')&&(ASCII<='9')) AS=ASCII-48;	

  for(color=0;color<3;color++)
  {
    if(color==0)        correctcol=Col_Green;
    else if(color==1)   correctcol=Col_Red;
    else if(color==2)   correctcol=Col_Blue;

    for (row=0;row<8;row++)
    {  
      if( (ASCII>='0')&&(ASCII<='9'))
        tempword=pgm_read_byte(&(ASCII_Number[AS][row]));  //like the monochromatic stuff...
      else
        tempword=pgm_read_byte(&(ASCII_Char[AS][row]));

      tempword=(shift<7)?(tempword<<shift):(tempword>>(shift-8)); //but with shifty stuff!

      for (dots=0;dots<4;dots++)
      {


        if((tempword<<(2*dots))&0x80)
        {
          tempdata&=0x0F;
          tempdata|=(correctcol<<4);
        }
        else
        {
          tempdata&=0x0F;
        }

        if((tempword<<(2*dots+1))&0x80)
        {
          tempdata&=0xF0;
          tempdata|=correctcol;
        }
        else
        {
          tempdata&=0xF0;
        }   


        dots_color[((Buffprt+1)&1)][color][row][dots]=tempdata;	  


      }
    }
  }
  Buffprt++;
  Buffprt&=1;

}


//===============RUNTIME=======================================
void RunCMD(void)
{
  switch (RainbowCMD[0][0]){  //figure out which function to run based on command
  case 'F':
    DispshowFrame();
    break;
    
  case 'T':
    DispshowTrich();
    break;
  
  case 'D':
    DispshowDich();
    break;

  case 'M':
    DispshowMono();
    break;

  case 'R':  //figure out which 'R' command function to run
    switch(RainbowCMD[0][1]){
    case showPrefabnicatel:
      DispshowPicture();	
      break;

    case showChar:
      DispshowChar();	
      break;

    case showColor:
      DispshowColor();	
      break;
    }
    break;
  }
}



//==============================================================
void shift_1_bit(unsigned char LS)  //shift 1 bit of  1 Byte color data into Shift register by clock
{
  if(LS)
  {
    shift_data_1;
  }
  else
  {
    shift_data_0;
  }
  clk_rising;
}
//==============================================================
void flash_next_line(unsigned char line,unsigned char level) // scan one line
{
  disable_oe;
  close_all_line;
  open_line(line);
  shift_24_bit(line,level);
  enable_oe;
}

//==============================================================
void shift_24_bit(unsigned char line,unsigned char level)   // display one line by the color level in buff
{
  unsigned char color=0,row=0;
  unsigned char data0=0,data1=0;
  le_high;
  for(color=0;color<3;color++)//GRB
  {
    for(row=0;row<4;row++)
    {
      data1=dots_color[Buffprt][color][line][row]&0x0f;
      data0=dots_color[Buffprt][color][line][row]>>4;

      if(data0>level)   //gray scale,0x0f aways light
      {
        shift_1_bit(1);
      }
      else
      {
        shift_1_bit(0);
      }

      if(data1>level)
      {
        shift_1_bit(1);
      }
      else
      {
        shift_1_bit(0);
      }
    }
  }
  le_low;
}



//==============================================================
void open_line(unsigned char line)     // open the scaning line 
{
  switch(line)
  {
  case 0:
    {
      open_line0;
      break;
    }
  case 1:
    {
      open_line1;
      break;
    }
  case 2:
    {
      open_line2;
      break;
    }
  case 3:
    {
      open_line3;
      break;
    }
  case 4:
    {
      open_line4;
      break;
    }
  case 5:
    {
      open_line5;
      break;
    }
  case 6:
    {
      open_line6;
      break;
    }
  case 7:
    {
      open_line7;
      break;
    }
  }
}



